/*
  Copyright (c) 2013 Simone Vacondio
  
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <LiquidCrystal.h>
#include <EEPROM.h>
#include <SdFat.h>
//#include <SdFatUtil.h>
//#include <math.h>

//button pins
#define OK_BUTTON 40
#define UPPER_BUTTON 48
#define LOWER_BUTTON 46
#define BACK_BUTTON 42

//lcd pins
#define LCD_RS 45
#define LCD_E 41
#define LCD_D4 31
#define LCD_D5 29
#define LCD_D6 27
#define LCD_D7 25
#define LCD_BLIGHT 22

//alarm pin
#define ALARM_PIN 36

//SD CS pin
#define SD_CS 53

/*EEPROM addresses
 EEPROM addresses usage:
 -0: number of calibration points;
 -1 to 63: pH and analog values for calibration;
 
 -100: number of titration curve points;
 -101 to 900: pH and mL values for titration curve;
 
 -1000: number of measure over time points;
 -1001: number of measure over time points;
 -1002: sample period in seconds;
 -1003: to 4000: pH values
 
 -4094: number used when saving a file with titration data to SD
 -4095: number used when saving a file with measure over time data
 to SD
 */
#define CALIB_N_ADDR 0
#define CALIB_PTS_ADDR 1

#define TITR_N_ADDR 100
#define TITR_PTS_ADDR 101
#define TITR_MAX_POINTS 200

#define SAMPLE_N1_ADDR 1000
#define SAMPLE_N2_ADDR 1001
#define SAMPLE_PERIOD_ADDR 1002
#define SAMPLE_PTS_ADDR 1003
#define SAMPLE_MAX_POINTS 1500

#define SD_TITR_NAME_ADDR 4094
#define SD_SAMPLE_NAME_ADDR 4095

//ph holding variable
double ph;

//linear/manual calibration parameters
double temperature = 298.15;

//polynomial calibration parameters
boolean polyn = false;
double coeff[10];
int nCoeff;

//alarm activator
boolean alarm = false;
boolean upperLimit;
double phLimit;

//listing and control variables
int mainList = 0;
int mainControl = 0;
char ctrl = '1';

//time holder
unsigned long currentMillis;

//backlight switch
boolean backlight=true;

//SD variables
SdFat sd;
boolean SD_av;

//lcd object
LiquidCrystal lcd(LCD_RS, LCD_E, LCD_D4, LCD_D5, LCD_D6, LCD_D7);

void setup () {
  //sets max analog input to 2.56V
  analogReference (INTERNAL2V56);

  //serial monitor setup
  Serial.begin(9600);

  //button setup
  pinMode (OK_BUTTON, INPUT_PULLUP);
  pinMode (UPPER_BUTTON, INPUT_PULLUP);
  pinMode (LOWER_BUTTON, INPUT_PULLUP);
  pinMode (BACK_BUTTON, INPUT_PULLUP);

  //alarm setup
  pinMode (ALARM_PIN, OUTPUT);

  //lcd setup
  lcd.begin(16, 2);
  pinMode(LCD_BLIGHT, OUTPUT);
  digitalWrite(LCD_BLIGHT, HIGH);
  lcd.clear();

  //SD setup
  if (!sd.begin(SD_CS, SPI_HALF_SPEED)) {
    //sd.initErrorHalt();
    SD_av=false;

    lcd.print("SD unavailable");
    lcd.setCursor(0, 1);
    lcd.print("Using EEPROM");
    delay (2000);
    lcd.clear();
  }
  else {
    SD_av=true;

    lcd.print("Using SD");
    delay(2000);
    lcd.clear();
  }

  //Serial.print("Free Ram: ");
  //Serial.println(FreeRam());
}

void loop () {
  //main menu
  while (mainControl==0) {
    switch (bState()){
    case UPPER_BUTTON:
      if (mainList<7){
        mainList++;
        lcd.clear();
      }
      break;

    case LOWER_BUTTON:
      if (mainList>0) {
        mainList--;
        lcd.clear();
      }
      break;

    case OK_BUTTON:
      mainControl = mainList;
      break;

    case BACK_BUTTON:
      mainList=0;
      lcd.clear();
      break;
    }

    switch (mainList){
    case 0:
      //idle screen, repeatedly printing pH and mV
      ph = readPh();
      lcd.home();
      lcd.print("pH:");
      lcd.print (ph);
      lcd.print(" ");
      alarmCheck(ph);
      
      lcd.print("mV:");
      lcd.print(temperature*0.00019842*(7-ph)*1000);
      lcd.print("  ");
      

      /*lcd.setCursor(0, 1);
       	 *	lcd.print("temp.: ");
       	 *	lcd.print(temperature);
       	 *	lcd.print(" C");
       	 *	break;*/

      //prints the analog input as well (only for development)
      delay(50);
      lcd.setCursor(0, 1);
      lcd.print("A0:");
      lcd.print(analogRead(A0));
      delay(50);
      lcd.print(" A1:");
      lcd.print(analogRead(A1));
      lcd.print("   ");

      break;

    case 1:
      lcd.home();
      lcd.print("Calibration");
      break;

    case 2:
      lcd.home();
      lcd.print("Set temperature");
      break;

    case 3:
      lcd.home();
      lcd.print("Set alarm");
      break;

    case 4:
      lcd.home();
      lcd.print("Titration curve");
      break;

    case 5:
      lcd.home();
      lcd.print("Measure");
      lcd.setCursor(0, 1);
      lcd.print("over time");
      break;

    case 6:
      lcd.home();
      lcd.print("Download data");
      break;

    case 7:
      lcd.home();
      lcd.print("Backlight");
      break;
    }
  }


  //calibration session
  while(mainControl==1) {
    if (ctrl == '0') {
      ctrl = '1';
      mainControl = 0;
      mainList = 0;
      lcd.clear();
      break;
    }

    int temp_nPoints = 2;
    int temp_analog[10];
    double temp_ph[10];

    int digits[] = {
      0, 7, 0, 0                                };
    boolean choice = false;

    if (polyn==true) {
      lcd.clear();
      lcd.print("Disable calibr.?");
      while (ctrl=='1') {
        if (choice==false) {
          lcd.setCursor (0, 1);
          lcd.print("No ");
        }
        else {
          lcd.setCursor(0, 1);
          lcd.print("Yes");
        }

        switch (bState()) {
        case UPPER_BUTTON: 
          choice=!choice; 
          break;
        case LOWER_BUTTON: 
          choice=!choice; 
          break;
        case BACK_BUTTON:
          lcd.clear();
          ctrl = '0';
          break;

        case OK_BUTTON:
          ctrl = '0';
          if (choice==true) polyn = false;
          break;
        }
      }
    }

    lcd.clear();
    lcd.print("Load latest?");
    while (ctrl=='1') {

      if (choice==false) {
        lcd.setCursor (0, 1);
        lcd.print("No ");
      }
      else {
        lcd.setCursor(0, 1);
        lcd.print("Yes");
      }

      switch (bState()) {
      case UPPER_BUTTON: 
        choice=!choice; 
        break;
      case LOWER_BUTTON: 
        choice=!choice; 
        break;
      case BACK_BUTTON:
        lcd.clear();
        ctrl = '0';
        break;

      case OK_BUTTON:
        if (choice==false) ctrl = '2';
        else ctrl='7';
        break;
      }
    }

    lcd.clear();
    lcd.print("How many points?");
    while (ctrl=='2') {
      switch (bState()) {
      case UPPER_BUTTON:
        if (temp_nPoints<10) temp_nPoints++;
        break;

      case LOWER_BUTTON:
        if (temp_nPoints>2) temp_nPoints--;
        break;

      case BACK_BUTTON:
        lcd.clear();
        ctrl='1';
        break;

      case OK_BUTTON:
        ctrl = '3';
        break;
      }
      lcd.setCursor(0,1);
      lcd.print(temp_nPoints);
      lcd.print(' ');
    }

    lcd.clear();
    lcd.print ("Point");
    lcd.setCursor(8, 0);
    lcd.print("pH:");
    lcd.setCursor(0, 1);
    lcd.print("analog in: ");

    if(ctrl=='3') {
      for (int i=0; ; i++) {
        while(ctrl=='3') {
          lcd.setCursor(5, 0);
          lcd.print(i+1);
          lcd.print(' ');

          lcd.setCursor(11, 1);
          lcd.print(readAnalog(0));

          selectNumber(digits, 4, 2, 11, 0, &ctrl);
        }

        if(ctrl == 'f') {
          temp_analog[i] = analogRead(0);
          temp_ph[i] = composeNumber(digits, 4, 2);
          if (i == temp_nPoints-1) {
            ctrl = '4';
            break;
          }
          ctrl = '3';
          continue;
        }
        if (ctrl == 'b') {
          if (i==0) break;
          ctrl = '3';
          i-=2;
          continue;
        }
      }
      if (ctrl=='b') ctrl = '2';

      if (ctrl == '4') {
        Serial.println("Points matrix:");
        for (int i=0; i<temp_nPoints; i++) {
          Serial.print(temp_analog[i]);
          Serial.print('\t');
          Serial.println(temp_ph[i]);
        }
        ctrl = '5';
      }  
    }

    lcd.clear();
    lcd.print("Save to EEPROM?");
    while (ctrl=='5') {
      if (choice==false) {
        lcd.setCursor (0, 1);
        lcd.print("No ");
      }
      else {
        lcd.setCursor(0, 1);
        lcd.print("Yes");
      }

      switch (bState()) {
      case UPPER_BUTTON: 
        choice=!choice; 
        break;
      case LOWER_BUTTON: 
        choice=!choice; 
        break;
      case BACK_BUTTON:
        lcd.clear();

        ctrl = '1';
        break;

      case OK_BUTTON:
        if (choice==false) ctrl = '8';
        else ctrl='6';
      }

    }

    //saves points to EEPROM
    //Serial.print("Free ram: ");
    //Serial.println(FreeRam());
    if (ctrl == '6') {
      EEPROM.write(CALIB_N_ADDR, temp_nPoints);

      //saves analog values
      int diff;
      int addr = CALIB_PTS_ADDR;
      //int serAddr = 1;
      for (int i=0; i<temp_nPoints; i++) {
        diff = temp_analog[i];
        for (int j=0; j<5; j++) {
          if (diff >= 255) {
            EEPROM.write(addr, 255);
            diff = diff - 255;
            addr++;
          }
          else {
            EEPROM.write (addr, diff);
            diff = 0;
            addr++;
          }
        }

        //serial print
        int sum=0;
        for (int j=addr-5; j<addr; j++) {
          int a = EEPROM.read(j);
          sum+=a;
          Serial.print(j);
          Serial.print("->");
          Serial.print(a);
          Serial.print('\t');
        }
        Serial.print("sum=");
        Serial.println(sum);
      }

      //saves pH values
      for (int i=0; i<temp_nPoints; i++) {
        int intg = floor(temp_ph[i]);
        int dec = ceil((temp_ph[i]-intg)*100);

        int a;
        int b;
        EEPROM.write(addr, intg);
        a=EEPROM.read(addr);
        Serial.print(addr);
        Serial.print("->");
        Serial.print(a);
        Serial.print('\t');
        addr++;

        EEPROM.write(addr, dec);
        b=EEPROM.read(addr);
        Serial.print(addr);
        Serial.print("->");
        Serial.print(b);
        Serial.print("\tsum=");

        double sum=a+ (double)b/100;
        Serial.println(sum);
        addr++;
      }
      ctrl = '8';
    }

    //fills the arrays with data from EEPROM
    if (ctrl=='7') {
      temp_nPoints = EEPROM.read(0);

      //reads saved analog inputs
      int addr=1;
      for(int i=0; i<temp_nPoints; i++) {
        int sum=0;
        for(int j=0; j<5; j++) {
          sum+=EEPROM.read(addr);
          addr++;
        }
        temp_analog[i]=sum;

        Serial.print(i);
        Serial.print("->");
        Serial.println(temp_analog[i]);
      }

      //reads saved pHs
      for(int i=0; i<temp_nPoints; i++) {
        temp_ph[i] = EEPROM.read(addr) + (double)EEPROM.read(addr+1)/100;
        addr+=2;

        Serial.print(i);
        Serial.print("->");
        Serial.println(temp_ph[i]);
      }
      ctrl='8';
    }

    lcd.clear();
    lcd.print("Processing...");
    //the marvelous gauss' elimination algorithm for calculating coefficients
    if (ctrl=='8') {
      const int maxp = temp_nPoints;
      double matrix[maxp][maxp+1];

      //fills matrix with data
      for (int i=0; i<maxp; i++) {
        for (int j=0; j<maxp; j++) {
          matrix[i][j]=pow(temp_analog[i], j);
        }
        matrix[i][maxp]=temp_ph[i];
      }

      //prints matrix
      for (int k=0; k<maxp; k++) {
        for (int j=0; j<maxp+1; j++) {
          Serial.print(matrix[k][j]);
          Serial.print('\t');
        }
        Serial.print('\n');
      }
      Serial.print('\n');

      //gauss' elimination algorithm
      for (int i=0; i<=maxp-2; i++) {

        //if all the column is 0, skips a cycle
        if(matrix[i][i]==0) {
          int a=0;
          for (int j=i+1; j<=maxp-1; j++) {
            if (matrix[j][i]!=0) {
              a++;
              break;
            }
          }
          if (a==0) {
            Serial.println("skipping");
            continue;
          }

          Serial.println("swapping");
          //if only the first element is 0, swaps the row with another one
          for (int j=i+1; j<=maxp-1; j++) {
            if (matrix[j][i]!=0) {
              for (int k=i; k<=maxp; k++) {
                double buf = matrix[j][k];
                matrix[j][k] = matrix [i][k];
                matrix[i][k] = buf;
              }
              break;
            }
          }
        }


        //zeroes a non-zero element by summing its row to a proper multiple of the first row
        for (int j=i+1; j<=maxp-1; j++) {
          if (matrix[j][i]!=0) {
            double coeff = matrix[j][i]/matrix[i][i];
            for(int k=i; k<=maxp; k++) matrix[j][k]-=matrix[i][k]*coeff;
          }
        }

        //prints matrix
        for (int k=0; k<maxp; k++) {
          for (int j=0; j<maxp+1; j++) {
            Serial.print(matrix[k][j]);
            Serial.print('\t');
          }
          Serial.print('\n');
        }
        Serial.print('\n');
      }

      //inverted gauss' elimination algorithm
      for (int i=maxp-1; i>=1; i--) {
        if(matrix[i][i]==0) {
          int a=0;
          for (int j=i-1; j>=0; j--) {
            if (matrix[j][i]!=0) {
              a++;
              break;
            }
          }
          if (a==0) {
            Serial.println("skipping");
            continue;
          }

          Serial.println("swapping");

          for (int j=i-1; j>=0; j--) {
            if (matrix[j][i]!=0) {
              for (int k=i; k>=0; k--) {
                double buf = matrix[j][k];
                matrix[j][k] = matrix [i][k];
                matrix[i][k] = buf;
              }
              double buf = matrix[j][maxp];
              matrix[j][maxp] = matrix[i][maxp];
              matrix[i][maxp] = matrix[j][maxp];

              break;
            }
          }
        }

        for (int j=i-1; j>=0; j--) {
          if (matrix[j][i]!=0) {
            double coeff = matrix[j][i]/matrix[i][i];
            for(int k=i; k>=0; k--) matrix[j][k]-=matrix[i][k]*coeff;
            matrix[j][maxp]-=matrix[i][maxp]*coeff;
          }
        }
        //prints matrix
        for (int k=0; k<maxp; k++) {
          for (int j=0; j<maxp+1; j++) {
            Serial.print(matrix[k][j]);
            Serial.print('\t');
          }
          Serial.print('\n');
        }
        Serial.print('\n');
      }

      //reduces all the coefficients to 1
      boolean err=false;
      for (int i=0; i<=maxp-1; i++) {
        if(matrix[i][i]!=1 && matrix[i][i]!=0) {
          matrix[i][maxp]=matrix[i][maxp]/matrix[i][i];
          matrix[i][i]=matrix[i][i]/matrix[i][i];
        }
        else if(matrix[i][i]==0) {
          err=true;
          break;
        }
      }
      if (err==true) {
        Serial.println("division by zero detected");
        lcd.clear();
        lcd.print("unexpected");
        lcd.setCursor(0, 1);
        lcd.print("failure");
        delay(2000);

        ctrl = '1';
        mainControl = 0;
        mainList = 0;
        break;
      }

      //prints matrix
      for (int k=0; k<maxp; k++) {
        for (int j=0; j<maxp+1; j++) {
          Serial.print(matrix[k][j]);
          Serial.print('\t');
        }
        Serial.print('\n');
      }
      Serial.print('\n');

      //assigns to the coefficient matrix the results and prints equation
      for (int i=0; i<maxp; i++) coeff[i]=matrix[i][maxp];

      //prints coefficients
      Serial.print ("y = ");
      for (int i=0; i<maxp; i++) {     
        Serial.print (coeff[i], DEC);
        Serial.print("*analog^");
        Serial.print(i);
        Serial.print(" + ");
        Serial.print('\n');
      }
      polyn = true;
      nCoeff = temp_nPoints;
      ctrl = '0';
    }
  }



  //temperature setting session
  while(mainControl==2) {
    int digits[] = {
      2, 9, 8, 1, 5                                };

    lcd.clear();
    lcd.print("Set temperature:");

    while (ctrl == '1') selectNumber (digits, 5, 2, 0, 1, &ctrl);
    
    lcd.print(" K  ");

    if (ctrl == 'f') {
      temperature = composeNumber(digits, 5, 2);

      lcd.clear();
      ctrl = '1';
      mainList = 0;
      mainControl = 0;
    }

    if (ctrl == 'b') {
      lcd.clear();
      exit();
    }

  }

  //alarm setting session
  while (mainControl==3) {
    int digits[] = {
      0, 7, 0, 0                                };
    boolean choice=false;

    lcd.clear();
    lcd.print("Activate alarm?");

    while (ctrl=='1') {
      if (choice==false) {
        lcd.setCursor (0, 1);
        lcd.print("No ");
      }
      else{
        lcd.setCursor(0, 1);
        lcd.print("Yes");
      }

      switch (bState()) {
      case UPPER_BUTTON: 
        choice=!choice; 
        break;
      case LOWER_BUTTON: 
        choice=!choice; 
        break;
      case BACK_BUTTON:
        lcd.clear();
        alarm=false;
        mainControl = 0;
        mainList = 0;
        ctrl = '0';
        break;

      case OK_BUTTON:
        if (choice==false) {
          alarm=false;

          lcd.clear();
          mainList=0;
          mainControl=0;
          ctrl = '0';
        } 
        else {
          alarm=true;

          ctrl='2';
        }
      }
    }
    if (ctrl=='0') {
      ctrl = '1';
      break;
    }

    lcd.clear();
    lcd.print("Select limit:");
    while (ctrl=='2') {
      if (choice==true) {
        lcd.setCursor(0, 1);
        lcd.print("Upper");
      } 
      else {
        lcd.setCursor(0, 1);
        lcd.print("Lower");
      }

      switch (bState()) {
      case UPPER_BUTTON: 
        choice=!choice; 
        break;
      case LOWER_BUTTON: 
        choice=!choice; 
        break;
      case BACK_BUTTON:
        lcd.clear();
        ctrl = '1';
        break;

      case OK_BUTTON:
        if (choice==true) {
          upperLimit=true;
          ctrl='3';
        } 
        else {
          upperLimit=false;
          ctrl='3';
        }
      }
    }

    lcd.clear();
    lcd.print("Select ph limit:");
    while(ctrl=='3') selectNumber(digits, 4, 2, 0, 1, &ctrl);

    if (ctrl=='f') {
      phLimit = composeNumber(digits, 4, 2);
      ctrl = '1';
      mainList = 0;
      mainControl = 0;
    }

    if (ctrl=='b') ctrl = '2';
    lcd.clear();
  }

  //titration curve session
  if (mainControl==4) {
    int digits [] = {
      0, 0, 0                                };

    double ml=0;
    int nPoints=0;
    int addr=TITR_PTS_ADDR;

    double infl[3];
    int n_infl=0;
    boolean past_infl=false;

    double prev_deriv;
    double prev_ml;
    double prev_ph;

    ofstream logfile;
    char buf[80];
    char filename[] = "TITR000.CSV";
    if(SD_av==true) {
      int val = EEPROM.read(SD_TITR_NAME_ADDR);
      if(val!=255) EEPROM.write(SD_TITR_NAME_ADDR, val+1);
      else EEPROM.write(SD_TITR_NAME_ADDR, 0);

      int a, b, c;

      if (val!=0) {
        a = val/100;
        b = (val-a*100)/10;
        c= (val-b*10-a*100);

        filename[4] = '0' + a;
        filename[5] = '0' + b;
        filename[6] = '0' + c;
      }
      if(sd.exists(filename)) sd.remove(filename);
      logfile.open(filename);

      if (!logfile.is_open()) {
        SD_av = false;

        lcd.clear();
        lcd.print("Unexpected");
        lcd.setCursor(0, 1);
        lcd.print("failure");
        delay(2000);
        lcd.clear();
        lcd.print("Using EEPROM");
        delay(2000);
      }
    }
    if(SD_av==true) {
      obufstream bout (buf, sizeof(buf));
      bout << filename << "\nmL\tpH\n";
      logfile << buf << flush;
    }

    lcd.clear();
    lcd.print("ph: ");
    lcd.setCursor(0, 1);
    lcd.print("mL: ");

    while (1) {
      while (ctrl=='1') {
        ph = readPh();
        lcd.setCursor(4, 0);
        lcd.print(ph);
        selectNumber(digits, 3, 2, 4, 1, &ctrl);
        alarmCheck(ph);
      }


      if (ctrl=='f') {
        ml+=composeNumber(digits, 3, 2);

        if(SD_av==false){
          nPoints++;

          int intg;
          int dec;

          intg = floor(ml);
          dec = ceil((ml-intg)*100);
          EEPROM.write(addr, intg);
          addr++;
          EEPROM.write(addr, dec);
          addr++;

          intg = floor(ph);
          dec = ceil((ph-intg)*100);
          EEPROM.write(addr, intg);
          addr++;
          EEPROM.write(addr, dec);
          addr++;

          if(nPoints==TITR_MAX_POINTS) {
            EEPROM.write(TITR_N_ADDR, nPoints);

            lcd.clear();
            lcd.print("Max point number");
            lcd.setCursor(0,1);
            lcd.print("achieved");

            exit();
            break;
          }
        } 
        else {
          if (nPoints==0) {
            prev_ph=ph;
            prev_ml=ml;
            prev_deriv=0;
          }
          else if (n_infl<3) {
            double deriv=(ph-prev_ph)/(ml-prev_ml);
            if (past_infl==false && abs(deriv) < abs(prev_deriv)){
              infl[n_infl] = prev_ml;
              n_infl++;
              past_infl = true;
            }
            else if (past_infl==true && abs(deriv) > abs(prev_deriv)) past_infl=false;

            prev_deriv = deriv;
            prev_ph = ph;
            prev_ml = ml;
          }
          nPoints++;

          logfile << ml << '\t' << ph << endl << flush;
          //Serial.print("Free ram: ");
          //Serial.println(FreeRam());
          if (!logfile) {
            lcd.clear();
            lcd.print("Unexpected");
            lcd.setCursor(0, 1);
            lcd.print("failure");
            delay(2000);

            exit();
            break;
          }
        }
        lcd.clear();
        lcd.print("Saved");
        delay(1000);
        lcd.clear();
        lcd.print("ph: ");
        lcd.setCursor(0, 1);
        lcd.print("mL: ");

        ctrl='1';
      }

      if (ctrl=='b') {
        if (SD_av==false) {
          EEPROM.write(TITR_N_ADDR, nPoints);

          lcd.clear();
          lcd.print("Data saved");
          lcd.setCursor(0, 1);
          lcd.print("to EEPROM");
          delay(2000);
        } 
        else {
          obufstream bout(buf, sizeof(buf));
          bout << "Inflection point(s)\n";
          for (int i=0; i<n_infl; i++) {
            bout<<infl[i]<<endl;
          }
          logfile << buf << flush;
          logfile.close();

          lcd.clear();
          lcd.print("Data saved");
          lcd.setCursor(0, 1);
          lcd.print("to SD card");
          delay(2000);
        }

        lcd.clear();
        exit();
        break;
      }
    }
  }


  //measure over time session
  while (mainControl==5) {
    int digits[] = {
      0, 0, 0                };

    int period;
    int n=0;
    unsigned long int prevMillis;

    int nPoints=0;    
    int addr=SAMPLE_PTS_ADDR;

    ofstream logfile;
    char buf[80];
    char filename[] = "SAMP000.CSV";
    if(SD_av==true) {
      int val = EEPROM.read(SD_SAMPLE_NAME_ADDR);
      if(val!=255) EEPROM.write(SD_SAMPLE_NAME_ADDR, val+1);
      else EEPROM.write(SD_SAMPLE_NAME_ADDR, 0);

      int a, b, c;

      if (val!=0) {
        a = val/100;
        b = (val-a*100)/10;
        c= (val-b*10-a*100);

        filename[4] = '0' + a;
        filename[5] = '0' + b;
        filename[6] = '0' + c;
      }
      if(sd.exists(filename)) sd.remove(filename);
      logfile.open(filename);

      if (!logfile.is_open()) {
        SD_av = false;

        lcd.clear();
        lcd.print("Unexpected");
        lcd.setCursor(0, 1);
        lcd.print("failure");
        delay(2000);
        lcd.clear();
        lcd.print("Using EEPROM");
        delay(2000);
      }
    }
    if(SD_av==true) {
      obufstream bout (buf, sizeof(buf));
      bout << filename << "\nseconds\tpH\n";
      logfile << buf << flush;
    }

    lcd.clear();
    lcd.print("Sample period(s)");

    while (ctrl=='1') selectNumber(digits, 3, 0, 0, 1, &ctrl);
    delay(500);

    if (ctrl=='f') {
      period = composeNumber(digits, 3, 0);
      if(period==0) {
        lcd.clear();
        lcd.print("Please set a");
        lcd.setCursor(0, 1);
        lcd.print("non-zero period");
        delay(2000);
        lcd.clear();
        lcd.print("Sample period(s)");

        ctrl = '1';
      } else if(period>255 && SD_av==false) {
        lcd.clear();
        lcd.print("Too big to");
        lcd.setCursor(0, 1);
        lcd.print("save to EEPROM");
        delay(2000);
        lcd.clear();
        lcd.print("Sample period(s)");

        ctrl = '1';
      }
      else {
        if (SD_av==false) EEPROM.write(SAMPLE_PERIOD_ADDR, period);
        ctrl='2';
      }
    }

    if (ctrl=='b') {
      lcd.clear();

      exit();
    }

    long unsigned int startMillis=millis();
    
    lcd.clear();
    lcd.print("pH:");
    lcd.setCursor(0, 1);
    lcd.print("seconds:");
    
    prevMillis=0;
    
    while (ctrl=='2') {
      ph=readPh();
      alarmCheck(ph);

      currentMillis=millis();
      if (currentMillis-prevMillis>=1000){
        lcd.setCursor(9,1);
        lcd.print((currentMillis-startMillis)/1000);
      }
      if (currentMillis-prevMillis>=period*1000) {
        n++;
        prevMillis=currentMillis;
        
        lcd.setCursor(4, 0);
        lcd.print(ph);
        
        if (SD_av==true) {
          logfile << n*period << '\t' << ph << endl;
          if (!logfile) {
            lcd.clear();
            lcd.print("Unexpected");
            lcd.setCursor(0, 1);
            lcd.print("failure");
            delay(2000);

            exit();
            break;
          }

        } 
        else {
          if (n < SAMPLE_MAX_POINTS) {          
            int intg;
            int dec;

            intg = floor(ph);
            dec = ceil((ph-intg)*100);
            EEPROM.write(addr, intg);
            addr++;
            EEPROM.write(addr, dec);
            addr++;
          } 
          else {
            int a;
            int b;
            a = n/255;
            b = n-a;

            EEPROM.write(SAMPLE_N1_ADDR, a);
            EEPROM.write(SAMPLE_N2_ADDR, b);

            lcd.clear();
            lcd.print("Max point number");
            lcd.setCursor(0,1);
            lcd.print("achieved");
            delay(2000);
            lcd.clear();

            exit();
            break;
          }
        }
      }

      if (bState()) {
        if (SD_av==true) {
          logfile.close();
          
          lcd.clear();
          lcd.print("Data saved");
          lcd.setCursor(0, 1);
          lcd.print("to SD card");
          delay(2000);
        }
        else {
          int a;
          int b;
          a = n/255;
          b = n-a;

          EEPROM.write(SAMPLE_N1_ADDR, a);
          EEPROM.write(SAMPLE_N2_ADDR, b);
          
          lcd.clear();
          lcd.print("Data saved");
          lcd.setCursor(0, 1);
          lcd.print("to EEPROM");
          delay(2000);
        }
        
        lcd.clear();
        exit();
      }
    }
  }

  //prints EEPROM values on serial
  if (mainControl==6) {
    //Serial.begin();
    Serial.print("Titration curve\nmL\tpH\n");

    int n = EEPROM.read(TITR_N_ADDR);
    int addr = TITR_PTS_ADDR;
    for (int i=0; i<n; i++) {
      double ml = EEPROM.read(addr) + (double)EEPROM.read(addr+1)/100;
      addr+=2;
      double ph = EEPROM.read(addr) + (double)EEPROM.read(addr+1)/100;
      addr+=2;

      Serial.print(ml);
      Serial.print('\t');

      Serial.print(ph);
      Serial.print('\n');
    }

    Serial.print("\nMeasure over time\nseconds\tpH\n");

    n=EEPROM.read(SAMPLE_N1_ADDR)*255+EEPROM.read(SAMPLE_N2_ADDR);
    addr=SAMPLE_PTS_ADDR;

    int period = SAMPLE_PERIOD_ADDR;
    for(int i=1; i<n; i++){
      double ph = EEPROM.read(addr) + (double)EEPROM.read(addr+1)/100;
      addr+=2;

      Serial.print(i*period);
      Serial.print('\t');
      Serial.print(ph);
      Serial.print('\n');
    }
    lcd.clear();
    mainControl=0;
  }


  //backlight switch
  if (mainControl==7) {
    backlight = !backlight;
    if (backlight) digitalWrite (LCD_BLIGHT, HIGH);
    else digitalWrite (LCD_BLIGHT, LOW);
    lcd.clear();
    mainControl=0;
  }
}

/*allows to enter a number by giviving an array as an argument
 *  and specifying the number of digits, decimals, column and row
 *  where to print and a program control variable ('f' for forward,
 *  'b' for back); you can compose the number from the array with
 *  composeNumber().
 */
void selectNumber (int *p, int nDigits, int nDecs, int column, int row, char* c) 
{
  static int currDigit = 0;

  switch(bState()) {
  case LOWER_BUTTON:
    if (p[currDigit]-1>=0) p[currDigit] -= 1; 
    break;

  case UPPER_BUTTON:
    if (p[currDigit]+1<=9) p[currDigit] += 1; 
    break;

  case OK_BUTTON:
    if (currDigit!=nDigits-1) currDigit++;
    else *c = 'f';
    break;


  case BACK_BUTTON:
    if (currDigit!=0) currDigit--;
    else *c = 'b';
    break;
  }


  //timer for number blinking
  static unsigned long previousMillis;
  static boolean state;
  currentMillis = millis();

  if (currentMillis-previousMillis>500) {
    previousMillis = currentMillis;
    state = !state;

    if (state==false) {
      lcd.setCursor(column, row);
      for(int i=0; i<=nDigits-nDecs-1; i++) lcd.print(p[i]);
      lcd.print(".");
      for (int i=nDigits-nDecs; i<=nDigits-1; i++) lcd.print(p[i]);

      if (currDigit>=nDigits-nDecs) {
        lcd.setCursor(column+currDigit+1, row);
        lcd.write(" ");
      } 
      else {
        lcd.setCursor(column+currDigit, row);
        lcd.write(" ");
      }
    } 
    else {
      lcd.setCursor(column, row);
      for(int i=0; i<=nDigits-nDecs-1; i++) lcd.print(p[i]);
      lcd.print(".");
      for (int i=nDigits-nDecs; i<=nDigits-1; i++) lcd.print(p[i]);
    }
  }     
}

/*returns a double type number starting from an array of digits,
 *  giving the number of digits and of decimals as arguments.
 */
double composeNumber (int *p, int nDigits, int nDecs) {
  double num=0;

  for (int i=0; i<=nDigits-1; i++) num+=(double)p[i]*pow(10, 
  (nDigits-nDecs-1)-i);

  return num;

}

/*converts the signal to ph according to calibration parameters
 *  and returns it;
 */
double readPh() {
  static unsigned long previousMillis;
  unsigned long currentMillis = millis();

  static double prevPh;

  if (currentMillis-previousMillis>500) {
    double ph;
    int sig1 = 2*analogRead(A1);
    int sig2 = 1023;
    int ph1 = 7;
    int ph2 = 14;
    if (polyn==false) ph = (double)(14-7)*(analogRead(A0)-sig1)/(sig2-sig1) + 7;
    else {
      double sum = 0;
      double x = analogRead(A0);
      for(int i=0; i<nCoeff; i++) sum += coeff[i]*pow(x, i);
      ph = sum;
    }
    previousMillis = currentMillis;
    prevPh = ph;
    return ph;
  } 
  else return prevPh;
}

int readAnalog(int pin) {
  static unsigned long previousMillis;
  unsigned long currentMillis = millis();

  static double prevAnalog;

  if (currentMillis-previousMillis>500) {
    double analog = analogRead(pin);
    previousMillis = currentMillis;
    prevAnalog = analog;
    return analog;
  } 
  else return prevAnalog;
}
/*returns which button is being pressed (same numbers as pins)
 *  and returns 0 if it is being pressed without releasing it before.
 */
int bState () {
  static int prevStatus;

  if (digitalRead(UPPER_BUTTON)==LOW && prevStatus!=UPPER_BUTTON) {
    prevStatus = UPPER_BUTTON;
    return UPPER_BUTTON;

  } 
  else if (digitalRead(LOWER_BUTTON)==LOW && prevStatus!=LOWER_BUTTON){
    prevStatus = LOWER_BUTTON;
    return LOWER_BUTTON;

  } 
  else if (digitalRead(OK_BUTTON)==LOW && prevStatus!=OK_BUTTON){
    prevStatus = OK_BUTTON;
    return OK_BUTTON;

  } 
  else if (digitalRead(BACK_BUTTON)==LOW && prevStatus!=BACK_BUTTON){
    prevStatus = BACK_BUTTON;
    return BACK_BUTTON;

  } 
  else if (digitalRead(UPPER_BUTTON)==LOW && prevStatus==UPPER_BUTTON){
    return 0;

  } 
  else if (digitalRead(LOWER_BUTTON)==LOW && prevStatus==LOWER_BUTTON){
    return 0;

  } 
  else if (digitalRead(OK_BUTTON)==LOW && prevStatus==OK_BUTTON){
    return 0;

  } 
  else if (digitalRead(BACK_BUTTON)==LOW && prevStatus==BACK_BUTTON){
    return 0;
  }

  prevStatus = 0;
  return 0;
}

/*if alarm is activated, keeps the ph limits under control and
 *  beeps if they are crossed
 */
void alarmCheck(double ph) {
  static unsigned long previousMillis;
  static boolean beep=true;
  static int duration=0;
  currentMillis = millis();

  if (alarm==true) {
    if (upperLimit==true && ph > phLimit){
      if (beep==true) {
        lcd.setCursor(15, 0);
        lcd.print('*');
      } 
      else {
        lcd.setCursor(15, 0);
        lcd.print(' ');
      }

      if (currentMillis-previousMillis>500){
        previousMillis=currentMillis;
        beep = !beep;
        if(beep==true) tone(ALARM_PIN, 2400);
        else noTone(ALARM_PIN);
        duration++;
      }

      if (duration==10) {
        alarm=false;
        noTone(ALARM_PIN);
        duration=0;
      }
      return;
    }

    if (upperLimit==false && ph < phLimit) {
      if (beep==true) {
        lcd.setCursor(15, 0);
        lcd.print('*');
      } 
      else {
        lcd.setCursor(15, 0);
        lcd.print(' ');
      }

      if (currentMillis-previousMillis>500){
        previousMillis=currentMillis;
        beep = !beep;
        if(beep==true) tone(ALARM_PIN, 2400);
        else noTone(ALARM_PIN);
        duration++;
      }

      if (duration==11) {
        alarm=false;
        duration=0;
      }
      return;
    }
    lcd.setCursor(15, 0);
    lcd.print('*');
  }

}

//resets control variables
void exit() {
  ctrl='1';
  mainList=0;
  mainControl=0;
}



